@@ -1,7 +1,8 @@
-// $Id: field.cpp,v 1.30 2000/09/14 22:31:39 eldamitri Exp $
+// $Id: field.cpp,v 1.46 2002/07/31 13:45:18 t1mpy Exp $
 
 // id3lib: a C++ library for creating and manipulating id3v1/v2 tags
 // Copyright 1999, 2000  Scott Thomas Haug
+// Copyright 2002 Thijmen Klok (thijmen@id3lib.org)
 
 // This library is free software; you can redistribute it and/or modify it
 // under the terms of the GNU Library General Public License as published by
@@ -24,13 +25,14 @@
 // id3lib.  These files are distributed with id3lib at
 // http://download.sourceforge.net/id3lib/
 
-#include <string.h>
-#include "field.h"
-#include "utils.h"
 
-#if defined HAVE_CONFIG_H
-#include <config.h>
-#endif
+#include "field_impl.h"
+#include "id3/utils.h" // has <config.h> "id3/id3lib_streams.h" "id3/globals.h" "id3/id3lib_strings.h"
+#include "field_def.h"
+#include "frame_def.h"
+#include "readers.h"
+
+using namespace dami;
 
 // This is used for unimplemented frames so that their data is preserved when
 // parsing and rendering
@@ -659,7 +661,7 @@ static ID3_FieldDef ID3FD_Volume[] =
   },
   { ID3FN_NOFIELD }
 };
-#endif /* _UNEFINED_ */
+#endif /* _UNDEFINED_ */
 
 // **** Currently Implemented Frames
 // APIC  PIC  ID3FID_PICTURE           Attached picture
@@ -758,14 +760,14 @@ static  ID3_FrameDef ID3_FrameDefs[] =
   {ID3FID_CDID,              "MCI", "MCDI", false, false, ID3FD_Unimplemented, "Music CD identifier"},
   {ID3FID_MPEGLOOKUP,        "MLL", "MLLT", false, true,  ID3FD_Unimplemented, "MPEG location lookup table"},
   {ID3FID_OWNERSHIP,         ""   , "OWNE", false, false, ID3FD_Unimplemented, "Ownership frame"},
+  {ID3FID_PRIVATE,           ""   , "PRIV", false, false, ID3FD_Private,       "Private frame"},
   {ID3FID_PLAYCOUNTER,       "CNT", "PCNT", false, false, ID3FD_PlayCounter,   "Play counter"},
   {ID3FID_POPULARIMETER,     "POP", "POPM", false, false, ID3FD_Popularimeter, "Popularimeter"},
   {ID3FID_POSITIONSYNC,      ""   , "POSS", false, true,  ID3FD_Unimplemented, "Position synchronisation frame"},
-  {ID3FID_PRIVATE,           ""   , "PRIV", false, false, ID3FD_Private,       "Private frame"},
   {ID3FID_BUFFERSIZE,        "BUF", "RBUF", false, false, ID3FD_Unimplemented, "Recommended buffer size"},
   {ID3FID_VOLUMEADJ,         "RVA", "RVAD", false, true,  ID3FD_Unimplemented, "Relative volume adjustment"},
   {ID3FID_REVERB,            "REV", "RVRB", false, false, ID3FD_Unimplemented, "Reverb"},
-  {ID3FID_SYNCEDLYRICS,      "SLT", "SYLT", false, false, ID3FD_SyncLyrics,	"Synchronized lyric/text"},
+  {ID3FID_SYNCEDLYRICS,      "SLT", "SYLT", false, false, ID3FD_SyncLyrics,    "Synchronized lyric/text"},
   {ID3FID_SYNCEDTEMPO,       "STC", "SYTC", false, true,  ID3FD_Unimplemented, "Synchronized tempo codes"},
   {ID3FID_ALBUM,             "TAL", "TALB", false, false, ID3FD_Text,          "Album/Movie/Show title"},
   {ID3FID_BPM,               "TBP", "TBPM", false, false, ID3FD_Text,          "BPM (beats per minute)"},
@@ -777,11 +779,11 @@ static  ID3_FrameDef ID3_FrameDefs[] =
   {ID3FID_ENCODEDBY,         "TEN", "TENC", false, true,  ID3FD_Text,          "Encoded by"},
   {ID3FID_LYRICIST,          "TXT", "TEXT", false, false, ID3FD_Text,          "Lyricist/Text writer"},
   {ID3FID_FILETYPE,          "TFT", "TFLT", false, false, ID3FD_Text,          "File type"},
-  {ID3FID_INITIALKEY,        "TKE", "TKEY", false, false, ID3FD_Text,          "Initial key"},
   {ID3FID_TIME,              "TIM", "TIME", false, false, ID3FD_Text,          "Time"},
   {ID3FID_CONTENTGROUP,      "TT1", "TIT1", false, false, ID3FD_Text,          "Content group description"},
   {ID3FID_TITLE,             "TT2", "TIT2", false, false, ID3FD_Text,          "Title/songname/content description"},
   {ID3FID_SUBTITLE,          "TT3", "TIT3", false, false, ID3FD_Text,          "Subtitle/Description refinement"},
+  {ID3FID_INITIALKEY,        "TKE", "TKEY", false, false, ID3FD_Text,          "Initial key"},
   {ID3FID_LANGUAGE,          "TLA", "TLAN", false, false, ID3FD_Text,          "Language(s)"},
   {ID3FID_SONGLEN,           "TLE", "TLEN", false, true,  ID3FD_Text,          "Length"},
   {ID3FID_MEDIATYPE,         "TMT", "TMED", false, false, ID3FD_Text,          "Media type"},
@@ -823,7 +825,7 @@ static  ID3_FrameDef ID3_FrameDefs[] =
   {ID3FID_NOFRAME}
 };
   
-/** \class ID3_Field
+/** \class ID3_Field field.h id3/field.h
  ** \brief The representative class of an ID3v2 field.
  ** 
  ** As a general rule, you need never create an object of this type.  id3lib
@@ -839,10 +841,10 @@ static  ID3_FrameDef ID3_FrameDefs[] =
  ** 
  ** Both types of strings support the GetNumTextItems() method.
  ** 
- ** An ASCII string field supports the Get(char*, size_t, index_t)), 
+ ** An ASCII string field supports the Get(char*, size_t, size_t)), 
  ** Set(const char*), Add(const char*), and operator=(const char*) methods.
  ** 
- ** A Unicode field also supports Get(unicode_t*, size_t, index_t),
+ ** A Unicode field also supports Get(unicode_t*, size_t, size_t),
  ** Set(const unicode_t*), Add(const unicode_t*), and 
  ** operator=(const unicode_t*).  Without elaborating, the Unicode
  ** methods behave exactly the same as their ASCII counterparts, taking
@@ -874,82 +876,42 @@ static  ID3_FrameDef ID3_FrameDefs[] =
  ** if you only plan to generate 3.0 tags.
  ** 
  ** @author Dirk Mahoney
- ** @version $Id: field.cpp,v 1.30 2000/09/14 22:31:39 eldamitri Exp $
+ ** @version $Id: field.cpp,v 1.46 2002/07/31 13:45:18 t1mpy Exp $
  ** \sa ID3_Tag
  ** \sa ID3_Frame
  ** \sa ID3_Err 
  **/
 
-ID3_Field::ID3_Field()
+ID3_FieldImpl::ID3_FieldImpl()
   : _id(ID3FN_NOFIELD),
     _type(ID3FTY_INTEGER),
     _spec_begin(ID3V2_EARLIEST),
     _spec_end(ID3V2_LATEST),
     _flags(0),
     _changed(false),
-    _integer(0),
-    _fixed_length(0),
-    _bytes(0),
+    _fixed_size(0),
     _num_items(0),
     _enc(ID3TE_NONE)
 {
   this->Clear();
 }
 
-ID3_Field::ID3_Field(const ID3_FieldDef& def)
+ID3_FieldImpl::ID3_FieldImpl(const ID3_FieldDef& def)
   : _id(def._id),
     _type(def._type),
     _spec_begin(def._spec_begin),
     _spec_end(def._spec_end),
     _flags(def._flags),
     _changed(false),
-    _fixed_length(def._fixed_length),
-    _bytes(_fixed_length),
+    _fixed_size(def._fixed_size),
     _num_items(0),
     _enc((_type == ID3FTY_TEXTSTRING) ? ID3TE_ASCII : ID3TE_NONE)
 {
-  if (_type == ID3FTY_TEXTSTRING)
-  {
-    _ascii = NULL;
-    _chars = _fixed_length;
-  }
-  else if (_type == ID3FTY_BINARY)
-  {
-    _binary = NULL;
-  }
   this->Clear();
 }
 
-ID3_Field::~ID3_Field()
+ID3_FieldImpl::~ID3_FieldImpl()
 {
-  if (_type == ID3FTY_INTEGER)
-  {
-    _integer = 0;
-  }
-  else if (_type == ID3FTY_BINARY)
-  {
-    if (_binary != NULL)
-    {
-      delete [] _binary;
-    }
-  }
-  else if (_type == ID3FTY_TEXTSTRING)
-  {
-    if (this->GetEncoding() == ID3TE_UNICODE)
-    {
-      if (_unicode != NULL)
-      {
-        delete [] _unicode;
-      }
-    }
-    else if (this->GetEncoding() == ID3TE_ASCII)
-    {
-      if (_ascii != NULL)
-      {
-        delete [] _ascii;
-      }
-    }
-  }
 }
 
 /** Clears any data and frees any memory associated with the field
@@ -957,76 +919,36 @@ ID3_Field::~ID3_Field()
  ** \sa ID3_Tag::Clear()
  ** \sa ID3_Frame::Clear()
  **/
-void ID3_Field::Clear()
+void ID3_FieldImpl::Clear()
 {
-  if (_type == ID3FTY_TEXTSTRING)
-  {
-    _chars = _fixed_length;
-  }
-  else
-  {
-    _bytes = _fixed_length;
-  }
   switch (_type)
   {
     case ID3FTY_INTEGER:
     {
       _integer = 0;
-      if (_bytes == 0)
-      {
-        _bytes = sizeof(uint32);
-      }
       break;
     }
     case ID3FTY_BINARY:
     {
-      if (_binary != NULL)
+      _binary.erase();
+      if (_fixed_size > 0)
       {
-        delete [] _binary;
-      }
-      if (_bytes == 0)
-      {
-        _binary = NULL;
-      }
-      else
-      {
-        _binary = new uchar[_bytes];
-        ::memset(_binary, '\0', _bytes);
+        _binary.assign(_fixed_size, '\0');
       }
       break;
     }
     case ID3FTY_TEXTSTRING:
     {
-      if (this->GetEncoding() == ID3TE_UNICODE)
+      _text.erase();
+      if (_fixed_size > 0)
       {
-        if (_unicode != NULL)
+        if (this->GetEncoding() == ID3TE_UNICODE)
         {
-          delete [] _unicode;
+          _text.assign(_fixed_size * 2, '\0');
         }
-        if (_chars == 0)
+        else if (this->GetEncoding() == ID3TE_ASCII)
         {
-          _unicode = NULL;
-        }
-        else
-        {
-          _unicode = new unicode_t[_chars];
-          ::memset((void *)_unicode, '\0', _chars * 2);
-        }
-      }
-      else if (this->GetEncoding() == ID3TE_ASCII)
-      {
-        if (_ascii != NULL)
-        {
-          delete [] _ascii;
-        }
-        if (_chars == 0)
-        {
-          _ascii = NULL;
-        }
-        else
-        {
-          _ascii = new char[_chars];
-          ::memset(_ascii, '\0', _chars);
+          _text.assign(_fixed_size, '\0');
         }
       }
       break;
@@ -1042,7 +964,7 @@ void ID3_Field::Clear()
 }
 
 bool
-ID3_Field::HasChanged() const
+ID3_FieldImpl::HasChanged() const
 {
   return _changed;
 }
@@ -1051,23 +973,27 @@ ID3_Field::HasChanged() const
  ** \brief Returns the size of a field.
  ** 
  ** The value returned is dependent on the type of the field.  For ASCII
- ** strings, this returns the number of characters in the field, no including
+ ** strings, this returns the number of characters in the field, not including
  ** any NULL-terminator.  The same holds true for Unicode---it returns the
  ** number of characters in the field, not bytes, and this does not include
  ** the Unicode BOM, which isn't put in a Unicode string obtained by the
- ** Get(unicode_t*, size_t, index_t) method anyway.  For binary and
+ ** Get(unicode_t*, size_t, size_t) method anyway.  For binary and
  ** integer fields, this returns the number of bytes in the field.
  ** 
  ** \code
- **   size_t howBig = myFrame.Field(ID3FN_DATA).Size();
+ **   size_t howBig = myFrame.GetField(ID3FN_DATA)->Size();
  ** \endcode
  ** 
  ** \return The size of the field, either in bytes (for binary or integer
  **         fields) or characters (for strings).
  **/
 
-size_t ID3_Field::BinSize() const
+size_t ID3_FieldImpl::BinSize() const
 {
+  if (_fixed_size > 0)
+  {
+    return _fixed_size;
+  }
   size_t size = this->Size();
   if (_type == ID3FTY_TEXTSTRING)
   {
@@ -1088,73 +1014,68 @@ size_t ID3_Field::BinSize() const
   return size;
 }
 
-size_t ID3_Field::Size() const
+size_t ID3_FieldImpl::Size() const
 {
   size_t size = 0;
   // check to see if we are within the legal limit for this field 0 means
   // arbitrary length field
-  if (_fixed_length > 0)
+  if (_fixed_size > 0)
   {
-    size = _fixed_length;
+    size = _fixed_size;
+  }
+  else if (_type == ID3FTY_INTEGER)
+  {
+    size = sizeof(uint32);
   }
   else if (_type == ID3FTY_TEXTSTRING)
   {
-    size = _chars;
+    size = _text.size();
   }
   else
   {
-    size = _bytes;
+    size = _binary.size();
   }
 
   return size;
 }
 
-size_t ID3_Field::Parse(const uchar *buffer, size_t buffSize)
+bool ID3_FieldImpl::Parse(ID3_Reader& reader)
 {
-  size_t bytesUsed       = 0;
-  
+  bool success = false;
   switch (this->GetType())
   {
     case ID3FTY_INTEGER:
     {
-      bytesUsed = ParseInteger(buffer, buffSize);
+      success = this->ParseInteger(reader);
       break;
     }
         
     case ID3FTY_BINARY:
     {
-      bytesUsed = ParseBinary(buffer, buffSize);
+      success = this->ParseBinary(reader);
       break;
     }
         
     case ID3FTY_TEXTSTRING:
     {
-      if (this->GetEncoding() == ID3TE_UNICODE)
-      {
-        bytesUsed = ParseUnicodeString(buffer, buffSize);
-      }
-      else
-      {
-        bytesUsed = ParseASCIIString(buffer, buffSize);
-      }
+      success = this->ParseText(reader);
       break;
     }
 
     default:
     {
-      ID3_THROW(ID3E_UnknownFieldType);
+      ID3D_WARNING( "ID3_FieldImpl::Parse(): unknown field type" );
       break;
     }
   }
-  
-  return bytesUsed;
+  return success;
 }
 
 ID3_FrameDef* ID3_FindFrameDef(ID3_FrameID id)
 {
   ID3_FrameDef  *info   = NULL;
 
-  for (index_t cur = 0; ID3_FrameDefs[cur].eID != ID3FID_NOFRAME; cur++)
+  for (size_t cur = 0; ID3_FrameDefs[cur].eID != ID3FID_NOFRAME; cur++)
   {
     if (ID3_FrameDefs[cur].eID == id)
     {
@@ -1171,7 +1092,7 @@ ID3_FindFrameID(const char *id)
 {
   ID3_FrameID fid = ID3FID_NOFRAME;
   
-  for (index_t cur = 0; ID3_FrameDefs[cur].eID != ID3FID_NOFRAME; cur++)
+  for (size_t cur = 0; ID3_FrameDefs[cur].eID != ID3FID_NOFRAME; cur++)
   {
     if (((strcmp(ID3_FrameDefs[cur].sShortTextID, id) == 0) &&
          strlen(id) == 3) ||
@@ -1186,58 +1107,58 @@ ID3_FindFrameID(const char *id)
   return fid;
 }
 
-size_t ID3_Field::Render(uchar *buffer) const
+void ID3_FieldImpl::Render(ID3_Writer& writer) const
 {
-  size_t bytesUsed = 0;
-  
   switch (this->GetType()) 
   {
     case ID3FTY_INTEGER:
-      bytesUsed = RenderInteger(buffer);
+    {
+      RenderInteger(writer);
       break;
+    }
         
     case ID3FTY_BINARY:
-      bytesUsed = RenderBinary(buffer);
+    {
+      RenderBinary(writer);
       break;
+    }
         
     case ID3FTY_TEXTSTRING:
-      bytesUsed = RenderString(buffer);
+    {
+      RenderText(writer);
       break;
+    }
         
     default:
-      ID3_THROW (ID3E_UnknownFieldType);
+    {
+      ID3D_WARNING ( "ID3D_FieldImpl::Render(): unknown field type" );
       break;
+    }
   }
-    
-  return bytesUsed;
 }
 
 ID3_Field &
-ID3_Field::operator=( const ID3_Field &rhs )
+ID3_FieldImpl::operator=( const ID3_Field &rhs )
 {
-  if (this != &rhs && this->GetType() == rhs.GetType())
+  const ID3_FieldImpl* fld = (const ID3_FieldImpl*) &rhs;
+  if (this != &rhs && this->GetType() == fld->GetType())
   {
-    switch (rhs.GetType())
+    switch (fld->GetType())
     {
       case ID3FTY_INTEGER:
       {
-        *this = rhs.Get();
+        this->SetInteger(fld->GetInteger());
         break;
       }
       case ID3FTY_TEXTSTRING:
       {
-        if (rhs.GetEncoding() == ID3TE_UNICODE)
-        {
-          this->Set_i(rhs._unicode, rhs.Size());
-        }
-        else if (rhs.GetEncoding() == ID3TE_ASCII)
-        {
-          this->Set_i(rhs._ascii, rhs.Size());
-        }
+        this->SetEncoding(fld->GetEncoding());
+        this->SetText(fld->GetText());
+        break;
       }
       case ID3FTY_BINARY:
       {
-        this->Set(rhs._binary, rhs.Size());
+        this->SetBinary(fld->GetBinary());
         break;
       }
       default:
@@ -1249,33 +1170,125 @@ ID3_Field::operator=( const ID3_Field &rhs )
   return *this;
 }
 
-bool ID3_Field::SetEncoding(ID3_TextEnc enc)
+bool ID3_FieldImpl::SetEncoding(ID3_TextEnc enc)
 {
   bool changed = this->IsEncodable() && (enc != this->GetEncoding()) &&
     (ID3TE_NONE < enc && enc < ID3TE_NUMENCODINGS);
   if (changed)
   {
-    // REMEMBER: _unicode and _ascii are in a union!
-    size_t size = this->Size();
-    if (size > 0)
-    {
-      if (_enc == ID3TE_UNICODE && enc == ID3TE_ASCII)
-      {
-        unicode_t* unicode = _unicode;
-        _ascii = new char[size];
-        ucstombs(_ascii, unicode, size);
-        delete [] unicode;
-      }
-      else if (_enc == ID3TE_ASCII && enc == ID3TE_UNICODE)
-      {
-        char* ascii = _ascii;
-        _unicode = new unicode_t[size];
-        mbstoucs(_unicode, ascii, size);
-        delete [] ascii;
-      }
-    }
+    _text = convert(_text, _enc, enc);
     _enc = enc;
     _changed = true;
   }
   return changed;
 }
+
+/** \class ID3_FrameInfo field.h id3/field.h
+ ** \brief Provides information about the frame and field types supported by id3lib
+ ** 
+ ** You normally only need (at most) one instance of the ID3_FrameInfo.  It
+ ** has no member data -- only methods which provide information about the
+ ** frame types (and their component fields) supported by id3lib as defined
+ ** in field.cpp .
+ **
+ ** Usage is straightforward.  The following function uses ID3_FrameInfo
+ ** to display a summary of all the frames known to id3lib:
+ ** \code
+ ** 
+ ** void ShowKnownFrameInfo {
+ **   ID3_FrameInfo myFrameInfo;
+ **   for (int cur = ID3FID_NOFRAME+1; cur <= myFrameInfo.MaxFrameID(); cur ++)
+ **   { 
+ **     cout << "Short ID: " << myFrameInfo.ShortName(ID3_FrameID(cur)) <<
+ **     " Long ID: " << myFrameInfo.LongName(ID3_FrameID(cur)) <<
+ **     " Desription: " << myFrameInfo.Description(ID3_FrameID(cur)) << endl;
+ **   }
+ ** } 
+ ** \endcode 
+ **
+ ** Functions are also provided to glean more information about the individual
+ ** fields which make up any given frame type.  The following for() loop,
+ ** embedded into the previous for() loop would provide a raw look at such
+ ** information.  Realize, of course, that the field type is meaningless
+ ** when printed.  Only when it is taken in the context of the ID3_FieldType enum
+ ** does it take on any meaningful significance.
+ **
+ ** \code
+ **  for (int cur = ID3FID_NOFRAME+1; cur <= fi.MaxFrameID(); cur ++)
+ **  {
+ **    int numfields = fi.NumFields(ID3_FrameID(cur));
+ **
+ **    cout << "ID: " << fi.LongName(ID3_FrameID(cur)) <<
+ **    " FIELDS: " << numfields << endl;
+ **    for(int i=0;i<numfields;i++) {
+ **      cout << "TYPE: " << fi.FieldType(ID3_FrameID(cur),i) <<
+ **      " SIZE: " << fi.FieldSize(ID3_FrameID(cur),i) <<
+ **      " FLAGS: " << fi.FieldFlags(ID3_FrameID(cur),i) << endl;
+ **
+ **    }
+ **
+ **    cout << endl;
+ **
+ **  }
+ ** \endcode
+ **
+ ** @author Cedric Tefft
+ ** @version $Id: field.cpp,v 1.46 2002/07/31 13:45:18 t1mpy Exp $
+ **/
+
+
+char *ID3_FrameInfo::ShortName(ID3_FrameID frameid)
+{
+  if(frameid < ID3FID_LASTFRAMEID)
+    return ID3_FrameDefs[frameid-1].sShortTextID;
+  else
+    return NULL;
+}
+
+char *ID3_FrameInfo::LongName(ID3_FrameID frameid)
+{
+  if(frameid < ID3FID_LASTFRAMEID)
+    return ID3_FrameDefs[frameid-1].sLongTextID;
+  else
+    return NULL;
+}
+
+const char *ID3_FrameInfo::Description(ID3_FrameID frameid)
+{
+  if(frameid < ID3FID_LASTFRAMEID)
+    return ID3_FrameDefs[frameid-1].sDescription;
+  else
+    return NULL;
+}
+
+int ID3_FrameInfo::MaxFrameID()
+{
+  return ID3FID_LASTFRAMEID-1;
+}
+
+int ID3_FrameInfo::NumFields(ID3_FrameID frameid)
+{
+  int fieldnum=0;
+
+  while (ID3_FrameDefs[frameid-1].aeFieldDefs[fieldnum]._id != ID3FN_NOFIELD)
+  {
+    fieldnum++;
+  }
+  return fieldnum;
+}
+
+ID3_FieldType ID3_FrameInfo::FieldType(ID3_FrameID frameid, int fieldnum)
+{
+  return (ID3_FrameDefs[frameid-1].aeFieldDefs[fieldnum]._type);
+}
+
+size_t ID3_FrameInfo::FieldSize(ID3_FrameID frameid, int fieldnum)
+{
+  return (ID3_FrameDefs[frameid-1].aeFieldDefs[fieldnum]._fixed_size);
+}
+
+flags_t ID3_FrameInfo::FieldFlags(ID3_FrameID frameid, int fieldnum)
+{
+  return (ID3_FrameDefs[frameid-1].aeFieldDefs[fieldnum]._flags);
+}
+
